from __future__ import annotations

from datetime import date
from enum import Enum

import pytz
from pydantic import BaseModel, Extra, Field, StrictBool, StrictStr, validator
from pydantic.types import T

from merqube_client_lib.pydantic_types import HolidayCalendarSpec


class HolidayCalendar(BaseModel):
    class Config:
        extra = Extra.forbid

    mics: list[StrictStr] = Field(default_factory=list)
    swaps_monitor_codes: list[StrictStr] = Field(default_factory=list)


class DecrementFeeType(Enum):
    fixed = "fixed"
    percentage_pre = "percentage_pre"
    percentage_post = "percentage_post"


class ClientIndexConfigBase(BaseModel):
    class Config:
        extra = Extra.forbid

    base_date: date
    # not all indices have this eg a single stock tr indices is just shares * price on the base_date
    base_value: int | float | None = None
    bbg_ticker: StrictStr | None = None
    currency: StrictStr = "USD"
    description: StrictStr
    email_list: list[StrictStr] = Field(default_factory=list)
    holiday_calendar: HolidayCalendar | None
    is_intraday: StrictBool = False
    name: StrictStr
    namespace: StrictStr
    run_hour: int
    run_minute: int
    timezone: StrictStr = "US/Eastern"
    title: StrictStr

    # set by the client lib
    holiday_spec: HolidayCalendarSpec | None = None

    @validator("timezone")
    def is_valid_tz(cls: T, v: str) -> str:  # pylint: disable=no-self-argument
        if v not in pytz.all_timezones:
            raise ValueError(f"Invalid timezone string: {v}")
        return v

    @validator("run_hour")
    def is_valid_hour(cls: T, v: int) -> int:  # pylint: disable=no-self-argument
        if v < 0 or v > 23:
            raise ValueError(f"Invalid timezone string: {v}")
        return v

    @validator("run_minute")
    def is_valid_minute(cls: T, v: int) -> int:  # pylint: disable=no-self-argument
        if v < 0 or v > 59:
            raise ValueError(f"Invalid timezone string: {v}")
        return v


class CorporateActions(BaseModel):
    reinvest_dividends: StrictBool = True


class ClientSSTRConfig(ClientIndexConfigBase):
    """
    Single stock total return index
    """

    class Config:
        extra = Extra.forbid

    underlying_ric: StrictStr


class ClientDecrementConfig(ClientIndexConfigBase):
    """
    Decrement index on top of an SSTR
    """

    class Config:
        extra = Extra.forbid

    underlying_index_name: StrictStr
    fee_value: float
    fee_type: DecrementFeeType
    day_count_convention: StrictStr
    client_owned_underlying: bool = True


class ClientMultiEquityBasketConfig(ClientIndexConfigBase):
    """
    Equity Basket
    """

    class Config:
        extra = Extra.forbid

    constituents_csv_path: StrictStr
    corporate_actions: CorporateActions = Field(default_factory=CorporateActions)
    level_overrides_csv_path: StrictStr | None = None



"""
Base class for all Merqube API Clients
"""
import operator
from collections import abc
from typing import Any, Iterable, Optional, cast

import pandas as pd
from cachetools import TTLCache, cachedmethod

# import like this so monkeypatch works as expected
from merqube_client_lib import session
from merqube_client_lib.constants import DEFAULT_CACHE_TTL
from merqube_client_lib.pydantic_types import IndexDefinitionPatchPutGet as Index
from merqube_client_lib.pydantic_types import IndexDefinitionPost
from merqube_client_lib.session import MerqubeAPISession
from merqube_client_lib.types import Manifest, ManifestList
from merqube_client_lib.types.secapi import (
    AddlSecapiOptions,
    MappingTable,
    SecapiMetricDefinition,
    SecAPIRecordsResponse,
)
from merqube_client_lib.util import batch_post_payload, pydantic_to_dict


class _MerqubeApiClientBase:
    """
    base class that contains validation functions
    """

    def __init__(
        self,
        user_session: Optional[session.MerqubeAPISession] = None,
        token: Optional[str] = None,
        **session_kwargs: Any,
    ):
        self.session = user_session or session.get_merqube_session(token=token, **session_kwargs)

    def _collection_helper(
        self,
        *,
        url: str,
        query_options: dict[str, str | Iterable[str] | None] | None = None,
    ) -> ManifestList:
        """
        common function to /security metrics and definitions
        """
        options: dict[str, str | list[str]] = {}

        for qo, v in (query_options or {}).items():
            if v is not None:
                options[qo] = v if isinstance(v, str) else ",".join(v)

        return self.session.get_collection(url=url, options=options)


class _IndexAPIClient(_MerqubeApiClientBase):
    """
    Indexapi class that contains methods that deal with multiple indices, creation of indices etc
    """

    def get_index_defs(
        self, index_names: str | list[str] | None = None, include_nonprod: bool = False
    ) -> dict[str, Manifest]:
        """
        Get index definitions for specified names, or all permissioned indices as a dictionary with ids as keys and index definitions as values

        supports index_names:
        - XXX - single name
        - XXX,YYY - comma separated names
        - [XXX,YYY] - list of names

        Note that MerQube's "collection apis" never return a 404 - these are search apis, and will return an empty list if no results are found
        """

        endpoint = "/index"
        if include_nonprod:
            endpoint += "?type=all"

        if not index_names:
            all_indices = self.session.get_collection(endpoint)
            return {i["id"]: i for i in all_indices}

        names_arg = index_names if isinstance(index_names, str) else ",".join(index_names)

        prod_clause = "&type=all" if include_nonprod else ""
        url = f"/index?names={names_arg}{prod_clause}"
        res = self.session.get_collection(url)
        return {i["id"]: i for i in res}

    def get_index_manifest(self, index_name: str) -> Manifest:
        """
        Get the model for a given index
        """
        return cast(Manifest, self.session.get_collection_single(f"/index?name={index_name}"))

    def get_index_model(self, index_name: str) -> Index:
        """
        Get the model for a given index
        """
        as_dict = self.session.get_collection_single(f"/index?name={index_name}")
        return Index.parse_obj(as_dict)

    def index_post_model_from_existing(
        self, index_name: str, reset_specific_fields: bool = True
    ) -> IndexDefinitionPost:
        """
        Gets a model that can be posted to /index to create a new index from an existing index

        For safety reasons, by default, the name and namespace are reset.
        """
        source_dict = self.get_index_manifest(index_name)
        source_dict.pop("id")
        source_dict.pop("status")

        model = IndexDefinitionPost.parse_obj(source_dict)

        if not reset_specific_fields:
            return model

        model.name = "test"
        model.namespace = "test"

        return model

    def create_index(self, index_def: IndexDefinitionPost) -> dict[str, str]:
        """
        Create an index
        Returns a dictionary containing the id of the index and its related securities (index, intraday_index)

        TODO: examples and index templates to be added to this repo.
        """
        return cast(dict[str, str], self.session.post("/index", json=pydantic_to_dict(index_def)).json())

    def update_index(self, index_def: Index) -> dict[str, str]:
        """
        Update (full object replacement) an index
        """
        return cast(dict[str, str], self.session.put(f"/index/{index_def.id}", json=pydantic_to_dict(index_def)).json())

    def patch_index(self, index_id: str, index_updates: Manifest) -> dict[str, str]:
        """
        Patch an index - index_updates is a partial index manifest
        """
        return cast(dict[str, str], self.session.patch(f"/index/{index_id}", json=index_updates).json())

    def delete_index(self, index_id: str) -> dict[str, str]:
        """
        Delete an index
        """
        return cast(dict[str, str], self.session.delete(f"/index/{index_id}").json())

    def replace_target_portfolio(self, index_id: str, target_portfolio: dict[str, Any]) -> dict[str, Any]:
        """
        Target portfolios are a PUT, not a POST, because an index can only have one active portfolio at a given time.
        "Last one wins", and it is not defined to have two+ active portfolios for the same timestamp.

        Internally (server side) all the history is kept, but only the latest one is used , and only the latest one is retrievable via this client.
        That is, if you PUT port1, then port2, (either via direct API or this library) port2 is always the one that is read on the next index run.
        Moreover, future portfolios can be posted at any time, but wont be considered by the index until that day.
        """
        return cast(
            dict[str, str], self.session.put(f"/index/{index_id}/target_portfolio", json=target_portfolio).json()
        )


class _SecAPIClient(_MerqubeApiClientBase):
    """
    Secapi client class
    """

    def __init__(
        self,
        user_session: Optional[MerqubeAPISession] = None,
        token: Optional[str] = None,
        **session_kwargs: Any,
    ):
        super().__init__(user_session=user_session, token=token, **session_kwargs)

        self.type_cache = TTLCache(1, ttl=DEFAULT_CACHE_TTL)  # type: ignore

    def get_supported_secapi_types(self) -> list[dict[str, str]]:
        """
        Get the list of supported security types
        """
        return self.session.get_collection(url="/security")

    @cachedmethod(operator.attrgetter("type_cache"))
    def _validate_secapi_type(self, sec_type: str) -> None:
        """Validate security_type"""
        assert sec_type in (
            supported_types := [x["name"] for x in self.get_supported_secapi_types()]
        ), f"sec_type must be one of {supported_types}"

    def _validate_single(self, *, sec_type: str, sec_id: str | None = None, sec_name: str | None = None) -> None:
        """Validate the input for functions that query for a single security"""
        self._validate_secapi_type(sec_type=sec_type)
        assert sec_id or sec_name, "Must provide either sec_id or sec_name"
        assert not (sec_id and sec_name), "Must provide either sec_id or sec_name, not both"

    def _validate_multiple(
        self,
        *,
        sec_type: str,
        sec_names: str | Iterable[str] | None = None,
        sec_ids: str | Iterable[str] | None = None,
        metrics: str | Iterable[str] | None = None,
    ) -> None:
        """Validate the input for functions that query for one or more securities"""
        self._validate_secapi_type(sec_type=sec_type)

        assert not (sec_ids and sec_names), "Must provide either sec_ids or sec_names, not both"

        for param in [sec_ids, sec_names, metrics]:
            if param:
                assert isinstance(param, (str, abc.Iterable))

    def _validate_chunking_options(
        self,
        *,
        metrics: str | Iterable[str],
        sec_names: str | Iterable[str] | None = None,
        sec_ids: str | Iterable[str] | None = None,
        addl_options: AddlSecapiOptions | None = None,
        metrics_chunk_size: int | None = None,
        securities_chunk_size: int | None = None,
    ) -> None:
        """the user wants to chunk by either metrics or securities"""
        if addl_options and addl_options.get("raw") == "true":
            raise NotImplementedError("Chunking while raw=true is not currently implemented")

        if metrics_chunk_size is not None and securities_chunk_size is not None:
            raise NotImplementedError("Chunking by both metrics and securities is not currently implemented")

        if metrics_chunk_size is not None:
            if isinstance(metrics, str):
                # this is what we get for trying to be nice and allow anything (single str)
                raise ValueError("Cannot use chunk size when metrics is a single string")

            if metrics_chunk_size < 1:
                raise ValueError("metrics_chunk_size cannot be < 1")

        if securities_chunk_size is not None:
            if securities_chunk_size < 1:
                raise ValueError("securities_chunk_size cannot be < 1")
            if not sec_names and not sec_ids:
                raise ValueError(
                    "when specifying securities_chunk_size, either sec_names or sec_ids must be an iterable of string"
                )
            if isinstance(sec_names, str):
                raise ValueError("Cannot use chunk size when sec_names is a single string")

            if isinstance(sec_ids, str):
                raise ValueError("Cannot use chunk size when sec_ids is a single string")

    def _get_security_metrics_helper(
        self,
        *,
        sec_type: str,
        metrics: str | Iterable[str],
        sec_names: str | Iterable[str] | None = None,
        sec_ids: str | Iterable[str] | None = None,
        start_date: str | pd.Timestamp | None = None,
        end_date: str | pd.Timestamp | None = None,
        addl_options: AddlSecapiOptions | None = None,
    ) -> SecAPIRecordsResponse:
        """
        if sec_names and sec_ids are both []/None, it gets ALL securities.
        """

        metrics_list = [metrics] if isinstance(metrics, str) else list(metrics)

        # Join did not work correctly for KeyViews and Tuples are badly behaved, so we'll remap them to lists.
        sec_names_list = ([sec_names] if isinstance(sec_names, str) else list(sec_names)) if sec_names else []
        sec_ids_list = ([sec_ids] if isinstance(sec_ids, str) else list(sec_ids)) if sec_ids else []

        query_options = {
            "metrics": metrics_list,
            "names": sec_names_list,
            "ids": sec_ids_list,
            "start_date": pd.Timestamp(start_date).isoformat() if start_date else start_date,
            "end_date": pd.Timestamp(end_date).isoformat() if end_date else end_date,
        }
        if addl_options:
            query_options.update(addl_options)

        return self._collection_helper(url=f"/security/{sec_type}", query_options=query_options)

    def get_metrics_for_security(
        self,
        sec_type: str,
        sec_id: str | None = None,
        sec_name: str | None = None,
    ) -> list[SecapiMetricDefinition]:
        """
        Get the list of metrics that are currently available for a security
        Can query by id or name
        """
        self._validate_single(sec_type=sec_type, sec_id=sec_id, sec_name=sec_name)

        return self.session.get_collection(
            f"/security/{sec_type}/{sec_id}/metrics" if sec_id else f"/security/{sec_type}/metrics?name={sec_name}"
        )

    def get_security_definitions_mapping_table(
        self,
        sec_type: str,
        sec_names: str | Iterable[str] | None = None,
        sec_ids: str | Iterable[str] | None = None,
        addl_options: AddlSecapiOptions | None = None,
    ) -> MappingTable:
        """
        Lists defined (and permissioned) securities for a type.
        Optionally filters by a list of securities.
        Returns a mapping table; either name -> id, or id -> name
        """
        # this function can be called with no ids/names to get a list of all permissioned securities
        self._validate_secapi_type(sec_type=sec_type)

        query_options: dict[str, str | Iterable[str] | None] = {"names": sec_names, "ids": sec_ids}
        if addl_options is not None:
            query_options.update(addl_options)

        rec_data = self._collection_helper(
            url=f"/security/{sec_type}",
            query_options=query_options,
        )

        return {c["id"]: c["name"] for c in rec_data} if sec_ids else {c["name"]: c["id"] for c in rec_data}

    def get_security_metrics(
        self,
        sec_type: str,
        metrics: str | Iterable[str],  # Secapi doesnt support fetching all metrics yet; cant be None
        sec_names: str | Iterable[str] | None = None,
        sec_ids: str | Iterable[str] | None = None,
        start_date: str | pd.Timestamp | None = None,
        end_date: str | pd.Timestamp | None = None,
        addl_options: AddlSecapiOptions | None = None,
        # if the secapi value is JSON, by default, pandas will flatten it into a dotted namespace
        # set this to control max_level: https://pandas.pydata.org/docs/reference/api/pandas.json_normalize.html
        # 0 means "dont normalize any JSONs
        normalize_level: int | None = None,
        metrics_chunk_size: int | None = None,
        securities_chunk_size: int | None = None,
    ) -> pd.DataFrame:
        """fetch security metrics from the SecAPI"""

        # Input validation
        self._validate_multiple(
            sec_type=sec_type,
            sec_names=sec_names,
            sec_ids=sec_ids,
            metrics=metrics,
        )
        assert metrics is not None, "Metrics cannot be None"

        params = {
            "sec_type": sec_type,
            "metrics": metrics,
            "sec_names": sec_names,
            "sec_ids": sec_ids,
            "start_date": start_date,
            "end_date": end_date,
            "addl_options": addl_options,
        }

        if metrics_chunk_size is None and securities_chunk_size is None:
            # no chunking
            data = self._get_security_metrics_helper(**params)  # type: ignore
            return pd.json_normalize(data, max_level=normalize_level)

        self._validate_chunking_options(
            addl_options=addl_options,
            metrics=metrics,
            sec_names=sec_names,
            sec_ids=sec_ids,
            metrics_chunk_size=metrics_chunk_size,
            securities_chunk_size=securities_chunk_size,
        )

        dfs: list[pd.DataFrame] = []

        if metrics_chunk_size is not None:
            for chunk in batch_post_payload(rows=list(metrics), batch_size=metrics_chunk_size):
                params["metrics"] = chunk
                data = self._get_security_metrics_helper(**params)  # type: ignore
                df = pd.json_normalize(data, max_level=normalize_level)

                # when we chunk by metrics, we may be missing some becuase the secapi doesnt return it if its None for all records
                for m in metrics:
                    if m not in df:
                        df[m] = None
                dfs.append(df)

        elif securities_chunk_size is not None:
            if sec_names:
                for chunk in batch_post_payload(rows=list(sec_names), batch_size=securities_chunk_size):
                    params["sec_names"] = chunk
                    data = self._get_security_metrics_helper(**params)  # type: ignore
                    dfs.append(pd.json_normalize(data, max_level=normalize_level))
            elif sec_ids:
                for chunk in batch_post_payload(rows=list(sec_ids), batch_size=securities_chunk_size):
                    params["sec_ids"] = chunk
                    data = self._get_security_metrics_helper(**params)  # type: ignore
                    dfs.append(pd.json_normalize(data, max_level=normalize_level))

        # the groupbys below squashes
        # eff1 id1 m1=NAN m2=x
        # eff2 id1 m1=Y  m2=NAN
        # into
        # eff1 id1 m1=Y m2=x
        # also:
        # for chunked, we return a consistent sort order of id, eff_ts
        return (
            pd.concat(dfs)
            .groupby(["eff_ts", "id"])
            .last()
            .reset_index()
            .sort_values(["id", "eff_ts"])
            .reset_index()
            .drop("index", axis=1)
        )



"""
Secapi types
"""
from typing import Any

from typing_extensions import NotRequired, TypedDict


class SecapiMetricDefinition(TypedDict):
    """The definition of a metric"""

    name: str
    description: NotRequired[str]
    data_type: NotRequired[str]
    object_schema: NotRequired[dict[str, Any]]


SecAPIRecordsResponse = list[dict[str, Any]]
AddlSecapiOptions = dict[str, Any]
MappingTable = dict[str, str]



import json
import logging
from copy import deepcopy
from typing import Any, Type, cast

import pandas as pd
from pydantic import BaseModel

from merqube_client_lib.api_client.merqube_client import MerqubeAPIClient
from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.pydantic_types import (
    HolidayCalendarSpec,
    IdentifierUUIDPost,
    IdentifierUUIDRef,
    IndexDefinitionPost,
    IndexReport,
    IntradayPublishConfigBloombergTarget,
    IntradayPublishConfigTarget,
    Provider,
    Stage,
)
from merqube_client_lib.templates.equity_baskets.schema import ClientIndexConfigBase
from merqube_client_lib.types import TargetPortfoliosDates
from merqube_client_lib.util import freezable_utcnow, get_token, pydantic_to_dict

SPEC_KEYS = ["base_date"]
DATE_KEYS = ["base_date"]
TOP_LEVEL = ["namespace", "name", "title", "base_date", "description"]


logger = get_module_logger(__name__, level=logging.DEBUG)


def read_file(filename: str) -> list[Any]:
    """reads portfolio/overrides files"""
    return list(pd.read_csv(filename, float_precision="round_trip").to_dict(orient="index").values())


def log_index(index: IndexDefinitionPost) -> None:
    """
    Logs the resulting index
    """
    json_formatted_str = json.dumps(json.loads(index.json(exclude_none=True)), indent=4)
    logger.info("Index spec: \n" + json_formatted_str)


def get_index_info(config_file_path: str, model: Type[ClientIndexConfigBase]) -> ClientIndexConfigBase:
    """
    load index specific info from config file + validate the data
    """
    # validate:
    config = json.load(open(config_file_path, "r"))

    index_info = model.parse_obj(config)

    swaps_monitor_codes, calendar_identifiers = None, None
    if hc := getattr(index_info, "holiday_calendar", None):
        calendar_identifiers = [f"MIC:{mic}" for mic in hc.mics] or []
        swaps_monitor_codes = hc.swaps_monitor_codes or None

    if not swaps_monitor_codes and not calendar_identifiers:
        calendar_identifiers = ["MIC:XNYS"]  # default to nyse

    # TODO: we may have to allow configuration of the weekmask
    index_info.holiday_spec = HolidayCalendarSpec(
        calendar_identifiers=calendar_identifiers,
        swaps_monitor_codes=swaps_monitor_codes,
    )  # type: ignore

    return index_info


def get_inner_spec(template: IndexDefinitionPost) -> dict[str, Any]:
    """
    Gets the inner spec of the index
    """
    assert template.spec and template.spec.index_class_args and template.spec.index_class_args.get("spec")
    return cast(dict[str, Any], template.spec.index_class_args["spec"])


def load_template(
    template_name: str,
    config_file_path: str,
    model: Type[ClientIndexConfigBase],
) -> tuple[MerqubeAPIClient, IndexDefinitionPost, BaseModel, dict[str, Any]]:
    """
    Loads a template index model to edit and then create a new index from
    """
    index_info = get_index_info(config_file_path=config_file_path, model=model)
    token = get_token()

    client = MerqubeAPIClient(token=token)

    template = client.index_post_model_from_existing(template_name)

    return client, template, index_info, get_inner_spec(template)


def _configure_report(template: IndexDefinitionPost, email_list: list[str] | None = None) -> None:
    """
    confifugre index reports
    """
    updated_reports = []
    if not email_list or not template.run_configuration or not template.run_configuration.index_reports:
        logger.warning("Warning, index has no reports to configure or has not been given an email list")
        return

    for ind, report in enumerate(template.run_configuration.index_reports):
        updated_report = cast(IndexReport, deepcopy(report))
        assert updated_report.program_args
        diss_config = json.loads(updated_report.program_args["diss_config"].replace('\\"', '"'))

        if not (dc := diss_config.get("email")) or len(dc) > 1:
            logger.warning(f"Warning, report {ind} has no email config or has more than one config")
            continue

        diss_config["email"][0]["recipients"] = email_list
        updated_report.program_args["diss_config"] = json.dumps(diss_config).replace('"', '\\"')
        updated_reports.append(updated_report)
    template.run_configuration.index_reports = updated_reports


def configure_index(
    template: IndexDefinitionPost, index_info: ClientIndexConfigBase, inner_spec: dict[str, Any]
) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    set up the index from index info
    """
    for k in TOP_LEVEL:
        v = getattr(index_info, k)
        if k == "base_date":
            v = pd.Timestamp(v).strftime("%Y/%m/%d")
        setattr(template, k, v)

    inner_spec["index_id"] = index_info.name
    for k in SPEC_KEYS:
        inner_spec[k] = getattr(index_info, k)

    # some indices want "calendar" instead of "holiday_spec"
    inner_spec["holiday_spec"] = inner_spec["calendar"] = index_info.holiday_spec

    name = index_info.name
    bbg_ticker = index_info.bbg_ticker
    is_intraday = index_info.is_intraday

    # set the start date at which this will start running
    assert template.run_configuration and template.run_configuration.schedule

    # set runtime
    # make sure to produce the last EOD value immediately on create
    today = freezable_utcnow() - pd.Timedelta(days=5)  # pyright: ignore
    template.run_configuration.schedule.schedule_start = pd.Timestamp(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=index_info.run_hour,
        minute=index_info.run_minute,
        second=0,
    ).isoformat()
    template.run_configuration.tzinfo = index_info.timezone

    template.stage = Stage.prod
    template.run_configuration.job_enabled = True

    if index_info.is_intraday:
        assert template.intraday and template.intraday.publish_config
        template.intraday.enabled = True

    bbg_identifier_post = None
    if bbg_ticker:
        template.identifiers = [IdentifierUUIDRef(name=bbg_ticker, provider=Provider.bloomberg)]
        if is_intraday:
            assert template.intraday and template.intraday.publish_config and template.intraday.publish_config.__root__
            cast(list[IntradayPublishConfigTarget], template.intraday.publish_config.__root__["price_return"]).append(
                IntradayPublishConfigTarget(
                    __root__=IntradayPublishConfigBloombergTarget(target="bloomberg", active_time_ranges=None)
                )
            )

        bbg_identifier_post = IdentifierUUIDPost(
            name=bbg_ticker, index_name=name, namespace=index_info.namespace, ticker=bbg_ticker
        )

    _configure_report(template=template, email_list=index_info.email_list)

    return template, bbg_identifier_post


def create_index(
    client: MerqubeAPIClient,
    template: IndexDefinitionPost,
    index_info: ClientIndexConfigBase,
    inner_spec: dict[str, Any],
    prod_run: bool,
    initial_target_portfolios: TargetPortfoliosDates | None = None,
) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    Creates an index from a template
    """
    template, bbg_post = configure_index(template=template, index_info=index_info, inner_spec=inner_spec)

    log_index(template)

    if initial_target_portfolios:
        logger.info("Initial target portfolios:")
        for date, itp in initial_target_portfolios:
            formatted = json.dumps(pydantic_to_dict(itp), indent=4, sort_keys=True)
            logger.info(f"{date}: {formatted}")

    if not prod_run:
        logger.debug("Dry run; exiting without creating the index")
        return template, bbg_post

    # if we are posting to bloomberg, we need to create the identifier first
    if bbg_post:
        res = client.create_identifier(provider=Provider.bloomberg, identifier_post=bbg_post)
        logger.info(f"Created identifier: {res}")

    res = client.create_index(index_def=template)
    logger.info(f"Created index: {res}")
    new_id = res["id"]

    for date, itp in initial_target_portfolios or []:
        logger.info(f"Pushing target portfolio for {date}")
        client.replace_target_portfolio(index_id=new_id, target_portfolio=pydantic_to_dict(itp))

    return template, bbg_post



"""
Create an equity basket index
"""
import logging
from typing import cast

from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.templates.equity_baskets.schema import ClientSSTRConfig
from merqube_client_lib.templates.equity_baskets.util import create_index, load_template
from merqube_client_lib.types import CreateReturn

logger = get_module_logger(__name__, level=logging.DEBUG)


def create(config_file_path: str, prod_run: bool = False) -> CreateReturn:
    """
    Creates a new Equity Basket with multiple entries
    This class does not handle Corax.
    """
    client, template, index_info, inner_spec = load_template(
        template_name="SINGLE_STOCK_TR_TEMPLATE_VERSION_1", config_file_path=config_file_path, model=ClientSSTRConfig
    )

    index_info = cast(ClientSSTRConfig, index_info)

    inner_spec["portfolios"]["constituents"] = [
        {
            "date": index_info.base_date,
            "identifier": index_info.underlying_ric,
            "quantity": 1,
        }
    ]

    inner_spec["currency"] = index_info.currency

    post_template, ident_ppost = create_index(
        client=client, template=template, index_info=index_info, inner_spec=inner_spec, prod_run=prod_run
    )

    return CreateReturn(post_template, ident_ppost)
